"""Shortcut to FluidLab interface classes (:mod:`pymanip.interfaces`)
=====================================================================

This module auto-imports all FluidLab interface classes.

"""

from .fluidlab_interfaces import *
