"""Andor module (:mod:`pymanip.video.andor`)
============================================

This module is a shortcut for the :class:`pymanip.video.andor.camera.Andor_Camera`
class.

"""

from .camera import Andor_Camera
