"""Concrete implementation with arduino ino file (:mod:`pymanip.aiodaq.arduino`)
================================================================================

This is just a convenient placeholder, since all methods are defined in
:mod:`pymanip.aioinstruments.arduino_gbf`.

"""

from pymanip.aioinstruments.arduino_gbf import AsyncArduino as ArduinoSystem
