"""Ximea module (:mod:`pymanip.video.ximea`)
========================================

This module is a shortcut for the :class:`pymanip.video.ximea.camera.Ximea_Camera`
class.

"""

from .camera import Ximea_Camera