"""Photometrics module (:mod:`pymanip.video.photometrics`)
==========================================================

This module is a shortcut for the :class:`pymanip.video.photometrics.camera.Photometrics_Camera`
class.

"""

from .camera import Photometrics_Camera